/*

	k-LinReg: an algorithm for switched linear regression
	
	Solves 
		minimize_w  sum_i min_j (y_i - w_j^T x_i)^2  
		
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <fenv.h>

#define INF HUGE_VAL
#define MAXFIRSTLINELENGTH 80

/* 
	Definition of a structure for the data
*/ 
struct datatype {
	double *X;
	double *y;
	int N;
	int n;
	int p;	
};

enum StoppingCriterionType {None, WeakCondition, StrongCondition};
struct Options {
	double epsilon2;
	double epsilon1;
	double epsilonINF;
	enum StoppingCriterionType NumRestarts;
	enum StoppingCriterionType TolMSE;
	enum StoppingCriterionType TolMAE;
	enum StoppingCriterionType TolINF;
	enum StoppingCriterionType TimeLimit;
	
	int MaxIterLocal;
	int Verbose;
};

/*
	Define th_data argument for working thread
*/
struct th_data { 
	int run;
	double *X;
	double *y;
	int N;
	int n;
	int p;

	double *w0; 	// initial vector
	double *w; 	// local minimizer on output
	double *fval;  	// local minimum
	double *MeanAE; // mean absolute error
	double *MaxAE;	// max absolute error
	// general settings:	
	int itermax;
	int verbose;
};

static pthread_mutex_t thread_data_mutex = PTHREAD_MUTEX_INITIALIZER;

// general STOP command (can be activated by Ctrl+C)
static int STOP = 0;
// TIMER
static int TIMER_TYPE = 0;
static int TIME_ALERT = 0;

/*
	Handler of ctrl+C signal
*/
void ctrlc_handler(int sig) {

	if(STOP == 1)
		exit(1);
	else {
		printf(" ***  Stopped by user ***\n");
		STOP = 1;
	}
}

/*
	Handler of ALARM signal (for time limit)
*/
void alarm_handler(int sig) {
	
	if(TIMER_TYPE == 1)
		TIME_ALERT = 1;
	else {
		printf(" ***  Time limit reached. ***\n");
		STOP = 1;
	}
}

/*
	Ex.: i = check_argv_eq(argc,argv,"-f") 
	
	Check if any argument of the command line is "-f"
	and returns the corresponding index of argv. 
	Returns 0 if "-f" cannot be found. 
*/
int check_argv_eq(int argc, char *argv[], char *str) {
	int i;
	int arg_str = 0;
	for(i = 1; i < argc; i++)
		if(strcmp(argv[i], str) == 0 ) {
			arg_str = i;
			break;
		}
		
	return arg_str;
}

/* Read data from file (and allocate memory)
	returns number of data points
	
	call : n = read_data("data.dat", &d, &X, &Y);
*/
int read_data(const char *filename, int *pd, double **X, double **Y, double *normalizeY, int logscale) {
	int n, d,i,j;
	FILE *fd;
	char buf[MAXFIRSTLINELENGTH];
	double *Xtmp,*Ytmp, maxY, minY,meanY, divY;
		
		
	if((fd = fopen(filename, "r")) == NULL) {
		printf("Wrong filename, abording.\n");
		return 0;
	}
	if (fgets(buf, MAXFIRSTLINELENGTH, fd) == NULL) {
		fclose(fd);
		printf("Error in data file, abording.\n");
		return 0;
	}
	sscanf(buf,"%d%d", &n,&d);
	*pd = d;
	
	Xtmp = (double *)malloc(sizeof(double) * n * d);
	Ytmp = (double *)malloc(sizeof(double) * n);
	
	meanY = 0;
	minY = INF;
	maxY = -INF;
	for(i=0;i<n;i++) {
		for (j=0;j<d;j++)
			if (fscanf(fd,"%lf", &Xtmp[i*d+j]) == 0) {
				printf("Error found in file %s\n",filename); 
				fclose(fd);
				*X = Xtmp;
				*Y = Ytmp;
				return i;
			}
		if(fscanf(fd,"%lf", &Ytmp[i]) == 0) {
				printf("Error found in file %s\n",filename); 
				fclose(fd);
				*X = Xtmp;
				*Y = Ytmp;
				return i;
		}
		if(Ytmp[i] > maxY) 
			maxY = Ytmp[i];
		if(Ytmp[i] < minY)
			minY = Ytmp[i];
		meanY += Ytmp[i];
					
	}
	meanY /= (double)n;
	
	if(*normalizeY > 0) {
		if(fabs(maxY) > fabs(minY)) 
			divY = fabs(maxY) ;
		else
			divY = fabs(minY);
			
		divY /= 10;
		for(i=0;i<n;i++) {
//			Ytmp[i] = (Ytmp[i] - meanY) / divY;
			Ytmp[i] = Ytmp[i]  / divY;
		}
		printf("maxY = %lf, divY = %lf\n",maxY,divY);
		*normalizeY = divY;
	}
	else
		*normalizeY = 1.;
	
	if(logscale) {
		for(i=0;i<n;i++) {
			Ytmp[i] = log(Ytmp[i] + 1 );
		}
	}
	
	fclose(fd);
	*X = Xtmp;
	*Y = Ytmp;
	return n;
}


/*
	Draw a random initialization for parameter vectors
	
*/
void init_w(const int n, const int p, double *w0, int *n_w0) {
	unsigned int j,k,l;
	
	double w_range = 100; // w0 is within [-w_range, w_range]
	double *w = (double*)malloc(sizeof(double) * n * p);

	for(j=0;j<n*p;j++) 
		w[j] = w_range * (2*drand48() - 1); 
		
	// add new w0 to the list
	for(j=0;j<n*p;j++) {
		w0[(*n_w0) * n * p + j] = w[j];			
	}
	*n_w0 = *n_w0 + 1;
	
	free(w);
	return;	
}


/*
	Basic k-LinReg algorithm (working thread)
	
	Optimize from a given initialization
	
*/
void *klinreg_thread(void *th_data) {

	unsigned int i,j,k,l;

	// Recover thread parameters
	struct th_data *data = (struct th_data *)th_data;
	double *X = data->X;
	double *y = data->y;
	int N = data->N;
	double *wt = data->w;
	double *fval = data->fval;
	double *MeanAE = data->MeanAE;
	double *MaxAE = data->MaxAE;	
	int n = data->n;
	int p = data->p;
	int itermax = data->itermax;
	int verbose = data->verbose;

	int run = data->run;
	
	// copy w0 -> wt
	for(j=0;j<n;j++)
		for(k=0;k<p;k++)
			wt[j*p+k] = data->w0[j*p+k];
	
	// Release mutex on th_data (end of critical section that launches threads)
	pthread_mutex_unlock(&thread_data_mutex);


	// local variables
	double eij,min_e,sum_e,sum_e_prec,absolute_error,stopping_criterion; 
	int iter;
	double epsilon = 0.0;
	double eps_f = 0.00;
	double *w = (double *)malloc(sizeof(double) * n * p); // parameter vectors
	int *S = (int*)malloc(sizeof(int) * N);
	
	double *Xiw = (double *)malloc(sizeof(double) * n); 

	int *Ni = (int*)malloc(sizeof(int) * n);

	
	double *WORK_= NULL;
	double *SingularValues = (double *)malloc(sizeof(double) * p);
	
	/*
		LIWORK >= max(1, 3 * MINMN * NLVL + 11 * MINMN),
	       where MINMN = MIN( M,N )  = p 
		
		 NLVL = MAX( 0, INT( LOG_2( MIN( M,N )/(SMLSIZ+1) ) ) + 1 )
		SMLSIZ is returned by ILAENV and is equal to the maximum
	         size of the subproblems at the bottom of the computation
		 tree (usually about 25)

	*/			
	int LIWORK = 3 * p * (1 + log2((double)p/(double)(2+1))) + 11 * p;
	int *IWORK_= (int*)malloc(sizeof(int) *  LIWORK);

	int LWORK_prec = 0;
	
	double *A = (double *)malloc(sizeof(double) * N*n*p);
	if(A == NULL) {
		printf("ERROR allocating A\n");
		exit(0);
	}
	double *B = (double *)malloc(sizeof(double) * N*n);
	if(A == NULL) {
		printf("ERROR allocating B\n");
		exit(0);
	}	
		
	// inner loop
	sum_e = INF;
	iter = 0;
	do {
		iter++;
		// compute S[i] : estimated mode of point i
		sum_e_prec = sum_e;
		sum_e = 0;
		for(i=0;i<N;i++) {
			min_e = INF;
			for(j=0;j<n;j++) {				
				Xiw[j] = X[i*p] * wt[j*p];	
				for(k=1;k<p;k++) 			
					Xiw[j] += X[i*p+k] * wt[j*p+k];  	// jth model output
				eij = y[i] - Xiw[j];
				eij *= eij; 			// squared error
				if (eij < min_e) {
					S[i] = j;
					min_e = eij;		// minimum of error
				}
			}
			sum_e += min_e;
			
		}	
		if(verbose>1)
			printf("f = %lf\n",sum_e);


		for(j=0;j<n;j++) 
			Ni[j] = 0;
		
		// compute new w
		for(i=0;i<N;i++) {
			for(j=0;j<n;j++) {				
				if(S[i] == j) {
					// A should be column major
					for(k=0;k<p;k++)
						A[j * p * N + k*N + Ni[j]] = X[i*p + k];

					B[j * N + Ni[j]] = y[i];
					Ni[j]++;
				}
			}
		}
		// update parameters	
		stopping_criterion = 0;			
		for(j=0;j<n;j++) {

		    if(Ni[j] > p) {	// Stop if a mode has less than p points
		    
			for(k=0;k<p;k++) {
				// save old w
				w[j*p+k] = wt[j*p+k];
			 }
			 
			/*
			 Solve the least squares problem
			 
			  min ||B - Ax||_2
			
			A : M x N matrix
			B : M x NRHS matrix 
			
			with LAPACK routine DGELSD

			*/			
			
			int info;
			char TRANS_ = 'N';
			int NRHS_ = 1;

			int LWORK_ = -1; // query optimal LWORK size
			
			double LWORK_opt;	
			
			int M_,N_,LDA_,LDB_,LIWORK;
			M_ = Ni[j];
			N_ = p;
			LDA_ = N; 
			LDB_ = M_;
			
			double RANK_; // output by lapack
			double RCOND_ = -1; // negative for machine precision
			
			// query memory requirements
			dgelsd_(&M_,&N_, &NRHS_, &A[j*p*N], &LDA_, &B[j*N], &LDB_, SingularValues, &RCOND_, &RANK_, &LWORK_opt, &LWORK_, &LIWORK, &info );
			LWORK_ = (int)LWORK_opt;
			
			if(LWORK_ > LWORK_prec) {
				WORK_ = (double*)realloc(WORK_, sizeof(double) * LWORK_);	
			}
			LWORK_prec = LWORK_;
			
			//printf("solve LS: LWORK=%d .WORK=%p, LIWORK=%d, IWORK=%p, info = %d with j=%d, Ni = %d\n",LWORK_,WORK_,LIWORK,IWORK_,info,j,Ni[j]);
			
			// Solve 
			dgelsd_(&M_,&N_, &NRHS_, &A[j*p*N], &LDA_, &B[j*N], &LDB_, SingularValues, &RCOND_, &RANK_, WORK_, &LWORK_, IWORK_, &info );
			
			if (info != 0)
				printf("Warning in dgelsd (least squares solution for X_%d): info = %d\n",j,info);

			// Convergence test
			for(k=0;k<p;k++) {
				wt[j*p+k] = B[j*N+k];
				
				if(verbose>1)
					printf("%lf, ",wt[j*p+k]);
				// test convergence with ||wt - w||^2 < eps
				stopping_criterion += (wt[j*p + k]  - w[j*p + k]) * (wt[j*p + k] - w[j*p + k]);
			}
		}	   		
		else {
			stopping_criterion = -1;
			break;
		}	
	    }
	} while(stopping_criterion > epsilon && (sum_e_prec - sum_e) > eps_f && iter < itermax && !STOP);


	// compute cost function value at local solution
	if(stopping_criterion<0)
		*fval = INF;
	else {
		sum_e = 0;
		if(MeanAE != NULL) 
			*MeanAE = 0;
		if(MaxAE != NULL) 
			*MaxAE = 0;
			
		for(i=0;i<N;i++) {
			min_e = INF;
			for(j=0;j<n;j++) {
				eij = y[i];				// target output
				for(k=0;k<p;k++) 			// minus
					eij -= X[i*p+k] * wt[j*p+k];  	// jth model output	
				eij *= eij; 			// squared error
				if (eij < min_e) {
					min_e = eij;			// minimum of error^2
					absolute_error = sqrt(eij);					
				}
			}
			sum_e += min_e; 
			if(MeanAE != NULL) 
				*MeanAE += absolute_error;
			if(MaxAE != NULL && absolute_error > *MaxAE) 
				*MaxAE = absolute_error;
		}
		*fval = sum_e/N;
		if(MeanAE != NULL) 
			*MeanAE /= N; 
	}
	
	if(verbose) { 
		if (stopping_criterion < 0)
			printf(" %d\t %d  \t (dropped one mode)\t\t %e\n",run, iter, *fval);			
		else if (stopping_criterion <= epsilon)
			printf(" %d\t %d  \t (||wt+1 - wt||^2 <= %1.1e )\t %e\n",run, iter, epsilon, *fval);
		else if ( iter >= itermax)
			printf(" %d\t %d  \t (max nb of iterations)\t\t %e\n",run, iter, *fval);			
		else
			printf(" %d\t %d  \t (cost_t - cost_t+1 < %lf)\t %e\n",run, iter, eps_f, *fval);
	}
	
	
	free(w);
	free(A);
	free(B);
	free(WORK_);
	
	free(SingularValues);
	free(IWORK_);
	
	free(Xiw);
	free(S);
	free(Ni);
	pthread_exit(EXIT_SUCCESS);
}

/*
	Parallel k-LinReg algorithm

	returns best cost function value
*/
double klinreg_global(double *X, double *y, double *wb, const int N, const int n, const int p, int *restarts_, struct Options *options ) {

	unsigned int i,j,k,l;
	
	int restarts = *restarts_;
	const double epsilon2 = options->epsilon2;
	const double epsilon1 = options->epsilon1;
	const double epsilonINF = options->epsilonINF;		
	
	const int verbose = options->Verbose;
	
	time_t cputime,starttime = time(NULL);
	
	if(restarts < 1)
		restarts = 10000;
	
	double fbest = INF;
	double MeanAEbest = INF;
	double MaxAEbest = INF;
	
	int run = 0;
	int ForceContinue = 0;
	int CanStop = 0;
	int t;
	int dotstep = 10;
	
	double *w0 = (double*)malloc(sizeof(double) * n * p * restarts); // keep a list of w0
	if( w0 == NULL ) {
		printf("Not enough memory for %d initializations in dimension %d.\n",restarts,n*p);
		exit(1);
	}
	double *w = (double*)malloc(sizeof(double) * n * p * restarts); // list of local minimizers
	if( w == NULL ) {
		printf("Not enough memory for %d minimizers in dimension %d.\n",restarts,n*p);
		exit(1);
	}
	
	int n_w0 = 0;
	int n_initializations = 0;
	
	pthread_t *threads = (pthread_t*)malloc(sizeof(pthread_t) * restarts); // threads id
	void *status; // for pthread_join
	int rc; // return code of pthread functions
		
	struct th_data thread_data;
	size_t stacksize = 100000000;
	pthread_attr_t attr;
	pthread_attr_init(&attr);
   	pthread_attr_setstacksize (&attr, stacksize);

	int r,rundone,bestrun;
	double *fval = (double*)malloc(sizeof(double) * restarts); // stores the local minima
	double *MeanAE, *MaxAE;
	
	// prepare common data for all threads
	thread_data.X = X;
	thread_data.y = y;						
	thread_data.N = N;		
	thread_data.n = n; 
	thread_data.p = p;
	thread_data.itermax = options->MaxIterLocal;
	if(options->TolMAE != None) 
		 MeanAE = (double*)malloc(sizeof(double) * NUMTHREADS); // mean absolute error
	else
		thread_data.MeanAE = NULL;
		
	if(options->TolINF != None)
		 MaxAE = (double*)malloc(sizeof(double) * NUMTHREADS); // max absolute error	
	else
		thread_data.MaxAE = NULL;
		
	if(verbose>0)
		thread_data.verbose = verbose-1;
	else
		thread_data.verbose = 0;
		 
	
	// Init display
	
	if(verbose > 1) {
		printf("\n Run   \t #iterations  (convergence) \t\t Cost (MSE)\n");
		printf("--------------------------------------------------------------\n");
		
	}
	else if(verbose == 1) {
		printf("\n Run   \t Cost function value (MSE)      ");
		if(options->TolMAE != None)
			printf("   Mean Absolute Error    ");
		if(options->TolINF != None)
			printf("      Max Abs. Error ");
		
		printf("\n-----------------------------------------------------");
		if(options->TolMAE != None)
			printf("-------------------------");
		if(options->TolINF != None)
			printf("--------------------");
		printf("\n");
	}

	// initialize random number generator

	srand48(time(NULL));
	rundone = 0;
	
	do { 
		// Launch NUMTHREADS threads with different initializations
		do {
		
			// random initialization of the parameters		
			init_w(n, p, w0, &n_w0);
			if(n_w0 >= restarts) {
				n_initializations += n_w0;
				n_w0 = 0;
			}
			
			// Enters critical section (and wait for previous thread to finish reading parameters)
			pthread_mutex_lock(&thread_data_mutex); 
	
			// set parameters for the thread
			thread_data.run = run; // #thread
			thread_data.w0 = &w0[n * p * (run % restarts)]; // initial vector
			thread_data.w = &w[n * p * (run % restarts)]; // local minimizer on output 	
			thread_data.fval = &fval[(run % restarts)]; // local minimum for this initialization
			if(options->TolMAE != None) 
				thread_data.MeanAE = &MeanAE[run - rundone];
			if(options->TolINF != None) 
				thread_data.MaxAE = &MaxAE[run - rundone];

			// launch thread
			rc = pthread_create(&threads[(run % restarts)], &attr, klinreg_thread, (void *) &thread_data);	
			if(rc != 0) {
				printf("Cannot create thread: probably due to a lack of memory, try to reduce stacksize.\n");
				exit(1);
			}
			
			run++;
		} while((run-rundone) < NUMTHREADS && run != restarts  && STOP!=1 ); 
		
		if (run == restarts && restarts%2 != 0)
			dotstep++;
			
		// Handle the local minimum and minimizers found so far
		for(r=rundone;r<run;r++) {

			// Wait for threads to terminate
			rc = pthread_join(threads[(r % restarts)],&status);
			
			// Update MAE criteria
			if(options->TolMAE != None && MeanAE[r-rundone] < MeanAEbest)
				MeanAEbest = MeanAE[r - rundone];

			if(options->TolINF != None && MaxAE[r-rundone] < MaxAEbest)
				MaxAEbest = MaxAE[r - rundone];
												
			//  Look for best minimum so far
			if(fval[(r % restarts)] < fbest) {
				fbest = fval[(r % restarts)];
				bestrun = r;		
				cputime = time(NULL) - starttime;
				
				// Copy best minimizer to wb
				for(k=0;k < n*p ; k++) 
					wb[k] = w[n*p*(bestrun % restarts) + k];
										
				
				// Test strong conditions (must be satisfied before stopping)
				ForceContinue = ((options->TolMSE==StrongCondition) && (fbest > epsilon2)) || ( (options->TolMAE==StrongCondition) && (MeanAEbest > epsilon1) ) || ( (options->TolINF==StrongCondition) && (MaxAEbest > epsilonINF) );
				
				// Test weak conditions (stop if any) 
				CanStop = ((options->TolMSE==WeakCondition) && (fbest <= epsilon2)) || ( (options->TolMAE==WeakCondition) && (MeanAEbest <= epsilon1) ) || ( (options->TolINF==WeakCondition) && (MaxAEbest <= epsilonINF) );
				
				if(verbose>1) {
					printf("    *** new best cost function value = %e (tol is %1.2e) *** \n",fbest,epsilon2);	
				}
				else if(verbose) {
					printf("\r  #%d\t %e (tol is %1.2e) ",bestrun,fbest,epsilon2);
					if(options->TolMAE != None)
						printf("   %1.2e (tol is %1.2e)",MeanAEbest,epsilon1);
					if(options->TolINF != None)
						printf("   %1.2e (tol is %1.2e)",MaxAEbest,epsilonINF);
					printf("\n");				
				}							
			}
		}
		rundone = run;  
		
		if(verbose == 1 && run%dotstep == 0){
			printf(".");
			fflush(stdout);
		}
		
		ForceContinue = ForceContinue || ( (options->NumRestarts==StrongCondition) && (run < restarts) ) || ( (options->TimeLimit==StrongCondition) && !TIME_ALERT);
		
		CanStop = CanStop || ( (options->NumRestarts==WeakCondition) && (run >= restarts) ) ; 
		// Weak condition on time acts directly on STOP
		
		// Final stopping condition
		STOP = STOP || (CanStop && !ForceContinue);
		
		// Loop until all StrongConditions are met and one of the WeakConditions is satisfied
	} while (STOP != 1);
	
	n_initializations += n_w0;
	
	if(verbose)
		printf("\nDone after %d / %d initializations tried. \nBest minimum was found after %d restarts in less than %u seconds. \n",n_initializations,*restarts_,bestrun,1+(unsigned int)cputime);
	
	*restarts_ = n_initializations; // number of restarts really performed
		
				
	free(w);free(w0);free(fval);free(threads);
	if(options->TolMAE != None) 
		free(MeanAE);
	if(options->TolINF != None) 
		free(MaxAE);
		
	return fbest;
		
}
/*
	Help
*/
void print_help(void) {
	printf(" \n k-LinReg algorithm for switched linear regression \n");
	printf("  -- a parallel implementation by F. Lauer (2011)\n\n");
	printf("Usage:\n");
	printf("\t klinreg datafile number_of_modes [options]\n\n");
	printf("Options:\n");
	printf("  -p     : probability of failure\n");
	printf("           which automatically defines the number of restarts\n");
	printf("           (default is 0.001)\n");
	printf("  -d     : alternate model of the probability of success for dynamical systems\n");
	printf("  -o     : base name of output files .model and .pred \n");
	printf("           (default is 'out', which creates 'out.model' and 'out.pred')\n");
	printf("  -v 	 : sets the verbose level from -1 to 2 (default is 1)\n");
	printf("\nStopping criteria:\n");
	printf("  -r, -R : max number of restarts / random initializations\n");
	printf("           (automatically determined by default)\n");

	printf("  -e, -E : tolerance on model mean squared error (MSE) (default is 0.0)\n");
	printf("  -a, -A : tolerance on model mean absolute error (default is none)\n");
	printf("  -b, -B : bound on model absolute error (inf-norm) (default is none)\n");
	
	printf("  -t, -T : time limit in seconds (default is none)\n");
	printf("\nMultiple criteria can be used:\n");
	printf("  - lowercase options allow the algorithm to stop\n");
	printf("              as soon as the given number is reached;\n");
	printf("  - UPPERCASE options force the algorithm to reach \n");
	printf("              the given number before it terminates.\n");
	printf("\n The algorithm can be stopped at any time by 'Ctrl-C' (or the SIGINT signal).\n");
	printf("\n See the README file for more info and examples. \n\n");
	
	return;
}

/*
	main function
*/
int main(int argc, char *argv[]) {
	
	// Default options
	struct Options options;
	options.epsilon2 = 0.0;
	options.epsilon1 = 0.0;
	options.epsilonINF = 0.0;
	
	options.NumRestarts = WeakCondition;
	options.TolMSE = WeakCondition;	// Stops if error = 0 by default as cost fct is non-negative
	options.TolMAE = None;
	options.TolINF = None;
	options.TimeLimit = None;
	
	options.MaxIterLocal = 500;
	options.Verbose = 1;

	int restarts=100;
	unsigned int maxtime = 0;
	
	// other variables
	unsigned int i,j,k;
	double *X, *y,fbest,yhat,yb,eij,min_e,mean_e; 
	int n,p, S,label,classif_e;
	double K;
	int tau,T;
	int dynamic = 0;
	
	// Default max probability of failure
	double Pfail = 0.001;

	
	double Psuccess;
	double normalizeY = 0;
	
	FILE *fout; 	// output prediction file 
	FILE *fm;	// model file
	FILE *fstdout=NULL; // redirect stdout for verbose < 0
	
	char datafile[250];
	char modelfile[250] = "out.model";
	char predfile[250] = "out.pred";
	
	/*
		Parse command line
	*/ 	
	if(argc<3) {	
		print_help();
		exit(0);
	}
	
	if(check_argv_eq(argc,argv, "-h") || check_argv_eq(argc,argv, "-help") || check_argv_eq(argc,argv, "--help") ) {
		i=system("cat README");
		exit(0);
	}
	
	// Get data filename
	strcpy(datafile, argv[1]);
	
	// read number of modes
	n = atoi(argv[2]);	
	
	// set verbose level
	i = check_argv_eq(argc,argv, "-v");
	if(i) {
		options.Verbose = atoi(argv[i+1]);
		if(options.Verbose < 0)  {
			fstdout = freopen(".klinreg.log", "w", stdout);
			if(fstdout == NULL)
				printf("Cannot set verbose level < 0 (error in stdout redirection)\n");
		}
	}
	
	printf("\n");
	
	// Set filenames options
	i = check_argv_eq(argc,argv, "-o");
	if(i) {
		strcpy(modelfile, argv[i+1]);
		strcpy(predfile, argv[i+1]);
		strcat(modelfile,".model");
		strcat(predfile,".pred");	
	}
	
	// Set tolerance options
	i = check_argv_eq(argc,argv, "-p");
	if(i) {
		Pfail = atof(argv[i+1]);
	}
	i = check_argv_eq(argc,argv, "-d");
	if(i) {
		dynamic = 1;
	}
		
	i = check_argv_eq(argc,argv, "-E");
	if(i) {
		options.TolMSE = StrongCondition;
		options.epsilon2 = atof(argv[i+1]);
		printf("Tolerance on MSE set to epsilon2 = %1.2e\n",options.epsilon2);
		printf(" (will not stop before reaching this accuracy)\n");
	}
	else {
		i = check_argv_eq(argc,argv, "-e");
		if(i) {
			options.TolMSE = WeakCondition;
			options.epsilon2 = atof(argv[i+1]);
			printf("Tolerance on MSE set to epsilon2 = %1.2e\n",options.epsilon2);
		}	
	}
	i = check_argv_eq(argc,argv, "-A");
	if(i) {
		options.TolMAE = StrongCondition;
		options.epsilon1 = atof(argv[i+1]);
		printf("Tolerance on Mean Absolute Error set to epsilon1 = %1.2e\n",options.epsilon1);
		printf(" (will not stop before reaching this accuracy)\n");		
	}
	else {
		i = check_argv_eq(argc,argv, "-a");
		if(i) {
			options.TolMAE = WeakCondition;
			options.epsilon1 = atof(argv[i+1]);
			printf("Tolerance on MAE set to epsilon1 = %1.2e\n",options.epsilon1);
		}	
	}
	i = check_argv_eq(argc,argv, "-B");
	if(i) {
		options.TolINF = StrongCondition;
		options.epsilonINF = atof(argv[i+1]);
		printf("Tolerance on Max. Absolute Error set to epsilonINF = %1.2e\n",options.epsilonINF);
		printf(" (will not stop before reaching this accuracy)\n");		
	}
	else {
		i = check_argv_eq(argc,argv, "-b");
		if(i) {
			options.TolINF = WeakCondition;
			options.epsilonINF = atof(argv[i+1]);
			printf("Tolerance on Max. Absolute Error set to epsilonINF = %1.2e\n",options.epsilonINF);			
		}	
	}

	// Set time limit
	i = check_argv_eq(argc,argv, "-T");
	if(i) {
		options.TimeLimit = StrongCondition;
		TIMER_TYPE = 1;
		maxtime = (unsigned int)atoi(argv[i+1]);
		printf("Will not stop before %u seconds\n",maxtime);
	}
	else {
		i = check_argv_eq(argc,argv, "-t");
		if(i) {
			options.TimeLimit = WeakCondition;
			maxtime = (unsigned int)atoi(argv[i+1]);
			printf("Time limit set to %u seconds\n",maxtime);
			printf(" (this overrides any accuracy conditions)\n");
		}	
	}

	/*
		Read data file
	*/
	int N = read_data(datafile, &p, &X, &y, &normalizeY,0);
	if (N == 0) {
		printf("No data.\n");
		return 1;
	}
		
	/* 
	 Compute the optimal number of restarts (r^*) satsfying
	
		r >=  log Pfail / log(1 - Psuccess)
	
	 with the model of the probaibility of success

		Psuccess = K(1 - exp(-(N-tau) / T) )
	
	 where 
		K = 0.99; 
		T = 2.22 * 2^n * p;
		tau = (2^n * p)^(1.4) * 0.2;
		
	or, for dynamical systems (with -d flag) 
		K = 1.02 - 0.023 * n;
		T = 52 * sqrt(2^n * p ) - 220;
		tau = 1.93 * 2^n * p - 37;
	*/
	
	tau = p;
	for (i=0;i<n;i++)
		tau *= 2;	// tau = 2^n * p
	
	if(dynamic) {
		K = 1.02 - 0.023 * n;	
		T = (int)round(52 * sqrt((double)tau) - 220);
		tau = (int)round(1.93 * tau - 37);
	}
	else {
		K = 0.99; 
		T = (int)round(2.22 * tau);
		tau = (int)round(0.2 * pow((double)tau,1.4));
	}

	Psuccess = K*(1 - exp(-(N-tau) / T) );
	
	if(Psuccess <= 0){
		Psuccess = 0;
		printf("Estimated probability of success = 0 \n (N=%d ; tau = %d,  T = %d )\n",N,tau,T);
		restarts = 100;
	}
	else {
		printf("Estimated probability of success (for each restart) = %e \n (N=%d ; K = %1.4lf, tau = %d,  T = %d )\n", Psuccess,N,K,tau,T);
		restarts = (int)ceil( log(Pfail) / log(1 - Psuccess) );
	}
	if(restarts <1)
		restarts = 1;
		
	printf("Number of restarts set to r^* = %d (for Pfailure = %1.1e) \n",restarts,Pfail);	
		
	// Check command line for a given number of restarts
	i = check_argv_eq(argc,argv, "-R");
	if(i) {
		options.NumRestarts = StrongCondition;
		restarts = atoi(argv[i+1]);
		printf(" Minimal number of restarts set to %d by user\n",restarts);
	}
	else {
		i = check_argv_eq(argc,argv, "-r");
		if(i) {
			options.NumRestarts = WeakCondition;
			restarts = atoi(argv[i+1]);
			printf(" Number of restarts set to %d by user\n",restarts);
		}	
		else {
			if(restarts < NUMTHREADS) {
				restarts = NUMTHREADS;
				printf("Number of restarts set to %d to match the number of CPUs\n",restarts);
			}
			else
				printf("Number of restarts set to %d \n",restarts);
		}
	}
	
		
	double *w = (double *)malloc(sizeof(double) * n * p); // parameter vectors
	
	
	// Install handler of SIGINT (ctrl+C) signal
	struct sigaction newhandler,oldhandler;
	memset(&newhandler, 0, sizeof(newhandler));
	newhandler.sa_handler = &ctrlc_handler;
	sigaction(SIGINT, &newhandler, &oldhandler);

	// Set alarm to stop optimization after maxtime seconds
	struct sigaction alarmnewhandler,alarmoldhandler;
	if(maxtime > 0) {
		memset(&alarmnewhandler, 0, sizeof(alarmnewhandler));
		alarmnewhandler.sa_handler = &alarm_handler;
		sigaction(SIGALRM, &alarmnewhandler, &alarmoldhandler);
	
		alarm(maxtime);
	}	
	
	/* 
		Start to Optimize
	*/
	printf("\nStarting to optimize %d models with %d points in R^%d...\n ",n,N,p);
	fbest = klinreg_global(X,y,w,N,n,p,&restarts,&options);
	
	
	printf("\n *** Best cost function value : %lf *** \n\n",fbest);

	/* 
		Compute predictions on training data
	*/

	// prediction file
	if( ((fout = fopen(predfile, "w")) == NULL)  ){
		printf("Error in prediction file.\n");
		return 0;
	}

	
	// output prediction
	int verbose_out = options.Verbose;
	if (verbose_out>2)
		printf("\n target y \t  best yhat \t S[i]\n ----------------------------------------------\n");
	mean_e = 0;
	classif_e = 0;
	for(i=0;i<N;i++) {
		if (verbose_out>2)
			printf(" %lf\t ",y[i]);

		min_e = INF;
		for(j=0;j<n;j++) {
			yhat = X[i*p]*w[j*p];
			for(k=1;k<p;k++)
				yhat += X[i*p+k]*w[j*p+k];
			eij = (y[i]-yhat)*(y[i]-yhat);
			if(eij < min_e) {
				S = j;
				min_e = eij;
				yb = yhat;
			}
			fprintf(fout,"%lf ",yhat);
		}
		mean_e += min_e;
		
		if (verbose_out>2) {
			printf("%lf\t", yb);	
			printf("%d\n",S);
		}
		fprintf(fout,"%lf ",yb);		
		fprintf(fout,"%d\n",S);
		
	}
	fclose(fout);
	
	// Data approximation error
	mean_e /= N;
	printf("\n      Mean squared error = %lf\n",mean_e * normalizeY * normalizeY);
	printf(" Root mean squared error = %lf\n",sqrt(mean_e));

	printf("\n Predictions written to %s\n",predfile);
	
	// Parameter vectors
	if(p<10) {
		printf("\n Parameter vectors:\n");
		for(j=0;j<n;j++) {
			printf("w_%d = [ %lf",j, normalizeY * w[j*p]);
			for(k=1;k<p;k++)
				printf(", %lf",normalizeY * w[j*p+k]);
			printf("]\n");
		}	
	}
	fm = fopen(modelfile, "w");
	if(fm != NULL) {
		for(j=0;j<n;j++) {
			fprintf(fm,"%lf",normalizeY * w[j*p]);
			for(k=1;k<p;k++)
				fprintf(fm, " %lf",normalizeY * w[j*p+k]);
			fprintf(fm,"\n");
		}	
		printf("\n Parameter vectors written to %s\n",modelfile);
	}
	
	
	printf("\n");
	
	free(w);free(X);free(y);
	return 0;
}
